from typing import Callable


def model_validator(model: dict, available_models: list) -> None:
    if not isinstance(model, str):
        raise ValueError(f"model: must be a string - the model entered is not in this required form - {model}")

    if model not in available_models:
        raise ValueError(f"model: must be one of {available_models} - the model entered is not in this required form - {model}")


def model_standardizer(model_selection: dict, default_model: str, default_params: dict) -> dict:
    if not isinstance(model_selection, dict):
        raise TypeError(
            f"model_selection: must be a dictionary with unique key(s): model, params - the model_selection entered is not in this required form - {model_selection}"
        )
    keys = list(model_selection.keys())
    if len(keys) > 2:
        raise ValueError(
            f"model_selection: must have only two keys: model, params - the model_selection entered is not in this required form - {model_selection}"
        )
    if len(set(keys) - set(["model", "params"])) > 0:
        raise ValueError(
            f"model_selection: must have only two keys: model, params - the model_selection entered is not in this required form - {model_selection}"
        )
    if "model" not in keys:
        model_selection["model"] = default_model
    if "params" not in keys:
        model_selection["params"] = default_params
    if len(model_selection["params"]) == 0:
        model_selection["params"] = default_params
    return model_selection


def model_selection_validator(model_selection: dict, available_models: list, params_validator: Callable) -> None:
    model = model_selection["model"]
    params = model_selection["params"]

    model_validator(model, available_models)

    params_validator(model, params)


def base_model_setup(
    model_selection: dict,
    available_models: list,
    default_model: str,
    default_params: dict,
    params_validator: Callable,
) -> dict:
    model_selection = model_standardizer(model_selection, default_model, default_params)

    model_selection_validator(model_selection, available_models, params_validator)

    return model_selection



from krixik.utilities.validators.system import EXPIRE_TIME_MIN
from krixik.utilities.validators.system import EXPIRE_TIME_MAX


def expire_time_checker(expire_time: str) -> None:
    # check if expire_time is a int and not bool
    if isinstance(expire_time, bool):
        raise ValueError(f"input expire_time must be an int - INFO: input expire_time type: {type(expire_time)}")

    if not isinstance(expire_time, int):
        raise ValueError(f"input expire_time must be an int - INFO: input expire_time type: {type(expire_time)}")

    # check that expire_time falls with in EXPIRE_TIME_MIN and EXPIRE_TIME_MAX
    if expire_time < EXPIRE_TIME_MIN:
        raise ValueError(f"expire_time {expire_time} must be at least {EXPIRE_TIME_MIN} seconds in the future")
    if expire_time > EXPIRE_TIME_MAX:
        raise ValueError(f"expire_time {expire_time} must be at most {EXPIRE_TIME_MAX} seconds in the future")



import os
from docx import Document
from krixik.utilities.validators.system.base.utilities.decorators import (
    file_converters_input_check,
)
from krixik.utilities.validators.data.docx import is_size as is_size_docx
from krixik.utilities.validators.data.text import is_size as is_size_text
from krixik.utilities.utilities import vprint


def delete_file(*, file_path: str, exception: str) -> None:
    if file_path is not None:
        if os.path.exists(file_path):
            os.remove(file_path)
        raise ValueError(
            f"there seems to be a problem with your docx file - file conversion failed with exception {str(exception)} - please check your docx file and see the converter documentation - python-docx - for further information: https://python-docx.readthedocs.io/en/latest/"
        )


@file_converters_input_check
def extract(
    *,
    local_file_path: str,
    local_save_directory: str | None = None,
    verbose: bool = True,
) -> str | None:
    # try conversion
    convert_save_path = ""
    if local_file_path is not None and local_save_directory is not None:
        try:
            file_name = "krixik_converted_version_" + os.path.splitext(os.path.basename(local_file_path))[0]
            extension = ".txt"
            convert_save_path: str = os.path.join(local_save_directory, file_name + extension)

            # remove file if it already exists
            if os.path.exists(convert_save_path):
                os.remove(convert_save_path)

            # creating a document reader object
            document = Document(local_file_path)

            # loop over paragraphs and write to txt
            num = 0
            for p in document.paragraphs:
                # get next line
                line = p.text

                if num == 0:
                    with open(convert_save_path, "a", encoding="utf-8") as new_file:
                        new_file.write(line)

                if num > 0:
                    if os.path.exists(convert_save_path):
                        with open(convert_save_path, "a", encoding="utf-8") as new_file:
                            new_file.write(line)
                    else:
                        raise FileNotFoundError(f"file {convert_save_path} does not exist")
                num += 1

            # write to file
            vprint(
                f"SUCCESS: File conversion complete using python-docx, result saved to: {convert_save_path}",
                verbose=verbose,
            )

            # report size check
            vprint(
                "INFO: Checking that file size falls within acceptable parameters...",
                verbose=verbose,
            )
            is_size_text(local_file_path=convert_save_path)
            vprint("INFO:...success!", verbose=verbose)

            return convert_save_path
        except ValueError as ve:
            vprint(
                "INFO:...failure!  The converted (text) file does not fall within acceptable size parameters.",
                verbose=verbose,
            )
            delete_file(file_path=convert_save_path, exception=str(ve))
        except Exception as e:
            delete_file(file_path=convert_save_path, exception=str(e))
    else:
        raise ValueError("the input local_file_path and/or local_save_directory is null")


@file_converters_input_check
def convert(
    *,
    local_file_path: str | None,
    local_save_directory: str | None = None,
    verbose: bool = True,
) -> str:
    """docx converter - converts docx file to txt using the pyton docx library

    Parameters
    ----------
    local_file_path : str | None
        path to local file to convert
    local_save_directory : str | None
        local directory to save converted file
    verbose : bool, optional
        by default True

    Returns
    -------
    str | None
        if conversion successful returns path to converted file
    """

    if local_file_path is None:
        raise ValueError("invalid local_file_path - local_file_path cannot be None")

    # check if local_file_path is a docx and convert to text if so
    if "." + local_file_path.split(".")[-1] == ".docx":
        if local_save_directory is None:
            raise ValueError("invalid local_save_directory - local_save_directory cannot be None")

        vprint("INFO: Converting docx to text...", verbose=verbose)

        # check size and validity of input docx file
        is_size_docx(local_file_path=local_file_path)

        # convert docx to text
        new_local_file_path = extract(
            local_file_path=local_file_path,
            local_save_directory=local_save_directory,
            verbose=verbose,
        )

        return new_local_file_path
    return local_file_path



import yaml
from krixik.__base__ import library_base_dir

config_path = f"{library_base_dir}/utilities/validators/data/utilities/config.yml"


def get_config() -> dict:
    with open(config_path, "r") as file:
        data_config = yaml.safe_load(file)["data_types"]

    return data_config


def get_inverse_config() -> dict:
    with open(config_path, "r") as file:
        data_config = yaml.safe_load(file)["data_types"]

    inverse_data_config = {}
    for data_type, extensions in data_config.items():
        for extension in extensions:
            inverse_data_config[extension] = data_type

    return inverse_data_config


def check_inverse_config(ext: str) -> dict:
    with open(config_path, "r") as file:
        data_config = yaml.safe_load(file)["data_types"]
    available_data_types = list(data_config.keys())

    inverse_data_config = {}
    for data_type, extensions in data_config.items():
        for extension in extensions:
            inverse_data_config[extension] = data_type
    try:
        return inverse_data_config[ext]
    except KeyError:
        raise ValueError(f"extension '{ext}' does not map to any currently available data types: {available_data_types} ")


def get_allowable_data_types(data_type: str) -> dict:
    with open(config_path, "r") as file:
        data_config = yaml.safe_load(file)["data_types"]

    if data_type not in data_config:
        raise ValueError(f"data_type: must be one of {data_config} - the data_type entered is not in this required form - {data_type}")
    return data_config[data_type]


def get_all_allowable_extensions() -> list:
    with open(config_path, "r") as file:
        data_config = yaml.safe_load(file)["data_types"]
    all_extension_lists = [v for k, v in data_config.items()]
    all_extensions = []
    for extension_list in all_extension_lists:
        all_extensions += extension_list
    return all_extensions



def quantize_validator(params: dict) -> None:
    if "quantize" not in list(params.keys()):
        raise ValueError(f"params: must have a key named 'quantize' - the params entered are not in this required form - {params}")
    if not isinstance(params["quantize"], bool):
        raise TypeError(f"quantize: must be a boolean - the quantize entered is not in this required form - {quantize}")



from krixik.utilities.validators.system.base.file_names import (
    file_names_checker as file_names_checker_utility,
)
from krixik.utilities.validators.system.base.file_names import (
    individual_file_name_checker as individual_file_name_checker_utility,
)
from krixik.utilities.validators.system.data.json import valid_extensions


def individual_file_name_checker(file_name: str, phase: str = "other") -> None:
    individual_file_name_checker_utility(file_name, phase=phase, valid_extensions=valid_extensions)


def file_names_checker(file_names: list, phase: str = "other", verbose: bool = True) -> None:
    file_names_checker_utility(file_names, phase=phase, valid_extensions=valid_extensions, verbose=verbose)



import re


# define pattern for valid api_key
api_key_pattern = r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"


def is_valid_uuid_string(input_string: str) -> bool:
    return bool(re.match(api_key_pattern, input_string))


# define pattern for valid api_url
api_url_pattern = r"^(http|https)://[a-zA-Z0-9._%+-]+(?:\/[a-zA-Z0-9.-?%&=]*)?$"


def is_valid_api_url(input_string: str) -> bool:
    return bool(re.match(api_url_pattern, input_string))


def api_key_checker(api_key: str) -> None:
    # check if api_key is None
    if api_key is None:
        raise ValueError("you failed to authenticate - API key cannot be None")

    # type check api_key
    if not isinstance(api_key, str):
        raise TypeError("you failed to authenticate - invalid API key type - API key must be a string")

    # check that api_key matches pattern
    res = is_valid_uuid_string(api_key)
    if res is False:
        raise ValueError("invalid api_key: please check that this is a valid and correct api_key")


def api_url_checker(api_url: str) -> None:
    # check if api_url is None
    if api_url is None:
        raise ValueError("you failed to authenticate - API url cannot be None")

    # type check api_url
    if not isinstance(api_url, str):
        raise ValueError("you failed to authenticate - invalid API url type - API url must be a string")

    # check that api_url matches pattern
    res = is_valid_api_url(api_url)
    if res is False:
        raise ValueError("invalid api_url: please check that this is a valid and correct api_url")



def verbose_checker(verbose: bool | None = None) -> None:
    if verbose is None:
        raise TypeError("verbose must be a boolean")
    if not isinstance(verbose, bool):
        raise TypeError("verbose must be a boolean")



import os
from pydub import AudioSegment
from moviepy.editor import AudioFileClip
from pydub.exceptions import PydubException


def is_valid(local_file_path: str | None) -> None:
    if local_file_path is None:
        raise ValueError("input file is None")

    # check that local_file_path represents audio file
    try:
        duration = 100
        AudioSegment.from_file(local_file_path, format="mp3", duration=duration)
    except PydubException as e:
        raise ValueError(f"input file is not a valid audio file - {local_file_path} - see exception {e}")
    except FileNotFoundError:
        raise FileNotFoundError(f"The file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"Error reading audio: {e}")


def is_size(
    *,
    local_file_path: str,
    minimum_seconds: int = 1,
    maximum_seconds: int = 180,
    minimum_file_size: float = 0.00001,
    maximum_file_size: float = 3.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"audio extraction failed with exception {e}")

    def compute_duration(file_path: str) -> float:
        try:
            audio_clip = AudioFileClip(file_path)
            duration_in_seconds = round(audio_clip.duration, 1)
            return duration_in_seconds
        except Exception as e:
            raise ValueError(f"audio duration calculation failed with exception {e}")

    try:
        # check that local_file_path is valid
        is_valid(local_file_path)

        # compute file size
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size or file_size > maximum_file_size:
            raise ValueError(
                f"input file size is {file_size} megabytes: either less than {minimum_file_size} megabytes (current minimum size allowable) or greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path}"
            )

        # compute length of audio file in seconds
        audio_duration = compute_duration(local_file_path)

        # check that audio file is  greater than minimum_seconds and less than maximum_seconds
        if audio_duration < minimum_seconds or audio_duration > maximum_seconds:
            raise ValueError(
                f"audio file is {audio_duration} seconds - this is either less than {minimum_seconds} seconds (current minimum size allowable) or greater than {maximum_seconds} seconds (current maximum size allowable)"
            )
    except ValueError as ve:
        raise ve
    except Exception as e:
        raise ValueError(f"audio extraction failed with exception {e}")



def wait_for_process_checker(wait_for_process: bool | None = None) -> None:
    if wait_for_process is None:
        raise TypeError("wait_for_process must be a boolean")
    if not isinstance(wait_for_process, bool):
        raise TypeError("wait_for_process must be a boolean")



MAX_FILE_ID_COUNT = 500
MAX_FILE_NAME_COUNT = 500
MAX_SYMBOLIC_PATH_COUNT = 500
MAX_FILE_NAME_LENGTH = 64
SYMBOLIC_DIRECTORY_PATH_MAX_LENGTH = 128
TAGS_COUNT_MAX = 20
TAG_KEY_LENGTH_MAX = 32
TAG_VALUE_LENGTH_MAX = 32
MAX_FILES_COUNT = 10000
FILE_DESCRIPTION_MAX_LENGTH = 256
EXPIRE_TIME_MIN = 60
EXPIRE_TIME_MAX = 60 * 60 * 24 * 30
EXPIRE_TIME_DEFAULT = 60 * 30
EXPIRE_TIME_DEFAULT_SEARCH = 60 * 60 * 24 * 7



import os
from krixik.pipeline_builder.module import Module
from krixik.pipeline_builder.pipeline import CreatePipeline
from krixik.modules import available_modules
from krixik.__base__ import library_base_dir

save_directory = library_base_dir + "/pipeline_examples/single_module"


def generate():
    try:
        for file in os.listdir(save_directory):
            if file.endswith(".yml"):
                os.remove(os.path.join(save_directory, file))

        pipeline = CreatePipeline()
        for module_name in available_modules:
            module = Module(module_name)
            pipeline = CreatePipeline(name=f"{module_name}-pipeline", module_chain=[module])
            pipeline.save(config_path=f"{save_directory}/{module_name}.yml")

    except Exception as e:
        raise e



def sort_order_checker(sort_order: str) -> None:
    if sort_order is not None:
        # check that sort_order is a string
        if not isinstance(sort_order, str):
            raise TypeError(f"sort_order: must be a string - the sort_order entered is not in this required form - {sort_order}")

        if sort_order not in ["ascending", "descending", "global"]:
            raise ValueError(
                f"sort_order: must be one of the following strings: 'ascending', 'descending', or 'global (if using vector_search only)'.  The sort_order entered is not in this required form - {sort_order}"
            )



import os
import importlib
import yaml
from krixik.__base__ import library_base_dir

modules_directory = library_base_dir + "/modules"
not_modules = ["__pycache__", "utilities"]
available_modules = [
    name for name in os.listdir(modules_directory) if os.path.isdir(os.path.join(modules_directory, name)) and name not in not_modules
]
available_modules.sort()


def get_module_details(module_name: str) -> dict:
    if not isinstance(module_name, str):
        raise Exception(f"module_name must be a string, your input {module_name} is of type {type(module_name)}")

    if module_name not in available_modules:
        raise Exception(f"user defined module {module_name} does not exist, currently available modules: {available_modules}")

    module_config_path = library_base_dir + f"/modules/{module_name}/module.yml"
    io_module_path = f"krixik.modules.{module_name}.io"

    module_config = None
    input_data_example = None
    output_data_example = None

    with open(module_config_path, "r") as file:
        module_config = yaml.safe_load(file)

    io_module = importlib.import_module(io_module_path)

    input_data_structure = io_module.InputStructure if hasattr(io_module, "InputStructure") else None
    if input_data_structure is not None:
        input_data_example = input_data_structure().data_example if hasattr(input_data_structure, "data_example") else None

    output_data_structure = io_module.OutputStructure if hasattr(io_module, "OutputStructure") else None

    if output_data_structure is not None:
        output_data_example = output_data_structure().data_example if hasattr(output_data_structure, "data_example") else None

    module_package = {
        "module_config": module_config,
        "input_data_example": input_data_example,
        "output_data_example": output_data_example,
    }

    return module_package



import requests
import json
from krixik.utilities.tree_illustrator import show_symbolic_file_tree
from krixik.system_builder.functions import show_tree_endpoint
from krixik.utilities.utilities import vprint


def show_post(self, *, symbolic_directory_path: str, max_files: int = 1000, verbose: bool = True):
    if hasattr(self, "_KrixikBasePipeline__version"):
        version = self._KrixikBasePipeline__version
    elif hasattr(self, "_KrixikSearchPipeline__version"):
        version = self._KrixikSearchPipeline__version
    else:
        raise ValueError("version not found in self")

    try:
        if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
            api_key = self._KrixikBasePipeline__api_key
            api_url = self._KrixikBasePipeline__api_url
        else:
            raise ValueError("api_key and api_url not found in self")

        if not hasattr(self, "_KrixikBasePipeline__pipeline"):
            raise ValueError("pipeline not found in input object")

        if symbolic_directory_path is None:
            raise TypeError("you must specify a symbolic_directory_path or stump")

        # prep headers
        headers = {"Content-Type": "text/plain", "krixikApiKey": api_key}

        # prep payload data
        payload_data = {
            "pipeline": self._KrixikBasePipeline__pipeline,
            "symbolic_directory_path": symbolic_directory_path,
            "max_files": max_files,
            "version": version,
        }

        # make request
        response = requests.post(
            api_url + show_tree_endpoint,
            headers=headers,
            json=payload_data,
            timeout=60,
        )

        # unpack response
        results = json.loads(response.text)
        self.request_id = results["request_id"]
        if response.status_code == 200 or response.status_code == 400:
            status_code_dict = {"status_code": response.status_code}
            status_code_dict.update(results)
            return status_code_dict
        else:
            raise ValueError(f"show_tree failed with status code {response.status_code}")
    except requests.exceptions.HTTPError as e:
        raise requests.exceptions.HTTPError(f"show_tree failed with HTTPError {e}")
    except requests.exceptions.ConnectionError as e:
        raise requests.exceptions.ConnectionError(f"show_tree failed with ConnectionError {e}")
    except requests.exceptions.Timeout as e:
        raise requests.exceptions.Timeout(f"show_tree failed with Timeout {e}")
    except requests.exceptions.RequestException as e:
        raise requests.exceptions.RequestException(f"show_tree failed with request exception {e}")
    except Exception as e:
        raise ValueError(f"show_tree failed with request exception {e}")
    finally:
        self._reset_class_variables()


def show_illustration(self, *, results: dict):
    try:
        # unpack items
        items = results["items"]

        # show paths if items is not None
        if items is not None:
            show_symbolic_file_tree(items)
        else:
            print("no items to show")
    except Exception as e:
        raise ValueError(f"show_tree failed with request exception {e}")



import re
from krixik.utilities.validators.system import FILE_DESCRIPTION_MAX_LENGTH

# validity pattern for file tags key and value
tags_pattern = r'^[a-zA-Z0-9\s\-_.,;\'"!?]*$'


def contains_only_alphanumeric_and_alphanumeric(input_string: str) -> bool:
    return bool(re.match(tags_pattern, input_string))


def file_description_checker(file_description: str) -> None:
    if file_description is not None:
        # check that file_description is a string
        if not isinstance(file_description, str):
            raise TypeError(f"invalid file description: not a string - {file_description}")

        # check that file_description length is less than FILE_DESCRIPTION_MAX_LENGTH
        if len(file_description) > FILE_DESCRIPTION_MAX_LENGTH:
            raise ValueError(
                f"invalid file description: length is greater than 256 characters (current maximum length allowable) - {file_description}"
            )

        # check that file_description contains only alphanumeric and alphanumeric characters
        if not contains_only_alphanumeric_and_alphanumeric(file_description):
            raise TypeError(f"invalid file description: contains non-alphanumeric/punctuation characters - {file_description}")



def use_default_clean_options_checker(use_default_clean_options: bool) -> None:
    # check that use_default_clean_options is not None
    if use_default_clean_options is not None:
        # check that use_default_clean_options is a bool
        if not isinstance(use_default_clean_options, bool):
            raise TypeError("use_default_clean_options is not a bool")



import re


def is_valid_version_string(s: str) -> bool:
    pattern = re.compile(r"^\d+\.\d+\.\d+$")
    return bool(pattern.match(s))


def version_checker(version: str) -> None:
    if not isinstance(version, str):
        raise ValueError("version must be a string")
    if not is_valid_version_string(version):
        raise ValueError(f"version must be valid and appears d.d.d, not as input {version}")



from regex import P
import yaml
from krixik.__base__ import library_base_dir
import importlib
from krixik.modules import available_modules
from krixik.utilities.validators.data.utilities.read_config import get_allowable_data_types


class Module:
    def __init__(self, module_type: str) -> None:
        if not isinstance(module_type, str):
            raise ValueError("module_type must be a string")

        self.name = module_type

        # check validity of input module
        if self.name not in available_modules:
            raise Exception(f"user defined module {self.name} does not exist")

        # define config and io paths
        self.__module_config_path = library_base_dir + f"/modules/{self.name}/module.yml"
        self.__io_module_path = f"krixik.modules.{self.name}.io"

        # load in module config
        self.__module_config = None
        if self.name is not None:
            with open(self.__module_config_path, "r") as file:
                self.__module_config = yaml.safe_load(file)
        else:
            raise Exception(f"please select a valid module from the available options: {available_modules}")
        # attach permitted extensions to module config
        self.__module_config["module"]["input"]["permitted_extensions"] = get_allowable_data_types(self.__module_config["module"]["input"]["type"])
        self.__module_config["module"]["output"]["permitted_extensions"] = get_allowable_data_types(self.__module_config["module"]["output"]["type"])

        io_module = importlib.import_module(self.__io_module_path)
        self.__input_dataclass = io_module.InputStructure if hasattr(io_module, "InputStructure") else None
        if self.__input_dataclass is None:
            raise Exception(f"error loading in {self.name} io module {self.__io_module_path} - InputStructure not found")

        self.__input_structure = self.__input_dataclass().__dict__
        self.__input_example = self.__input_dataclass().data_example
        self.__input_format = self.__input_dataclass().format
        self.__input_process_key = self.__input_dataclass().process_key
        self.__input_process_type = self.__input_dataclass().process_type

        self.__output_dataclass = io_module.OutputStructure if hasattr(io_module, "OutputStructure") else None
        if self.__output_dataclass is None:
            raise Exception(f"error loading in {self.name} io module {self.__io_module_path} - OutputStructure not found")

        self.__output_structure = self.__output_dataclass().__dict__
        self.__output_example = self.__output_dataclass().data_example
        self.__output_format = self.__output_dataclass().format
        self.__output_process_key = self.__output_dataclass().process_key
        self.__output_process_type = self.__output_dataclass().process_type

    @property
    def config(self):
        return self.__module_config

    @property
    def input_structure(self):
        return self.__input_structure

    @property
    def output_structure(self):
        return self.__output_structure

    @property
    def input_format(self):
        return self.__input_format

    @property
    def output_format(self):
        return self.__output_format

    @property
    def input_example(self):
        return self.__input_example

    @property
    def output_example(self):
        return self.__output_example

    @property
    def input_process_key(self):
        return self.__input_process_key

    @property
    def output_process_key(self):
        return self.__output_process_key

    @property
    def input_process_type(self):
        return self.__input_process_type

    @property
    def output_process_type(self):
        return self.__output_process_type

    @property
    def click_data(self):
        return {
            "module_name": self.name,
            "input_format": self.input_format,
            "output_format": self.output_format,
            "input_process_key": self.input_process_key,
            "input_process_type": self.input_process_type,
            "output_process_key": self.output_process_key,
            "output_process_type": self.output_process_type,
        }



import os
import imghdr
from PIL import Image


def is_valid(local_file_path: str) -> bool | None:
    try:
        with open(local_file_path, "rb") as f:
            # Read the first 10 bytes to determine the image type
            header = f.read(10)
            f.seek(0)  # Reset the file pointer to the beginning

            # Use imghdr to determine the image type
            image_type = imghdr.what(None, header)

            # Check if the image type is JPEG or PNG
            if image_type in ["jpg", "jpeg", "png"]:
                return True
            else:
                raise ValueError(f"The file '{local_file_path}' does not represent a valid image.")
    except FileNotFoundError:
        raise FileNotFoundError(f"The file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"Error reading image: {e}")


def is_proportionally_acceptable(local_file_path: str) -> None:
    try:
        # Check if the image is valid
        is_valid(local_file_path)

        # Open the image file
        with Image.open(local_file_path) as img:
            width, height = img.size
            if width == 0 or height == 0:
                raise ValueError("image width or height is 0")

            if width / height < 0.05 or height / width < 0.05:
                raise ValueError("image proportion is less than 0.2: width/height or height/width")
    except ValueError as ve:
        raise ve
    except Exception as e:
        raise ValueError(f"invalid local_file_path: {e}")


def is_size(
    *,
    local_file_path: str,
    minimum_file_size: float = 0.001,
    maximum_file_size: float = 5.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"file size calculation failed with exception {e}")

    try:
        # check that local_file_path represents a valid json file
        is_valid(local_file_path)

        # check size of input json file
        file_size = compute_size(local_file_path)
        if file_size < minimum_file_size or file_size > maximum_file_size:
            raise ValueError(
                f"input file size is {file_size} megabytes: either less than {minimum_file_size} megabytes (current minimum size allowable) or greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path}"
            )
    except ValueError as ve:
        raise ve
    except Exception as e:
        raise ValueError(f"invalid local_file_path: {e}")

    # check proportion of input image file
    is_proportionally_acceptable(local_file_path)



CHUNK_SIZE_MIN = 3
CHUNK_SIZE_MAX = 20
OVERLAP_SIZE_MIN = 1


def fixed_params_validator(params: dict) -> None:
    params_key_options = ["chunk_size", "overlap_size"]
    if params is not None:
        if not isinstance(params, dict):
            raise TypeError(
                f"params: must be a dictionary with unique key(s): chunk_size, overlap_size - the params entered is not in this required form - {params}"
            )

        # if params is length 0, then return 1
        if len(params) == 0:
            raise ValueError(
                f"params: must be a dictionary with unique key(s): chunk_size, overlap_size - the params entered is not in this required form - {params}"
            )

        # check that only two keys are in params
        if len(params.keys()) > 2:
            raise ValueError(
                f"params: must be a dictionary with unique key(s): chunk_size, overlap_size - the params entered is not in this required form - {params}"
            )

        # check that params has all required keys
        for key in params.keys():
            if key not in params_key_options:
                raise ValueError(
                    f"params: must be a dictionary with unique key(s): chunk_size, overlap_size - the params entered is not in this required form - {params}"
                )

        # check that params values are all ints
        for value in params.values():
            if not isinstance(value, int) or isinstance(value, bool):
                raise TypeError(f"params: all values must be integers - the params entered is not in this required form - {params}")

        # check that no key is repeated
        if len(params.keys()) != len(set(params.keys())):
            raise ValueError(
                f"params: must be a dictionary with unique key(s): chunk_size, overlap_size - the params entered is not in this required form - {params}"
            )

        # check that chunk_size is an integer
        if "chunk_size" in params.keys():
            if not isinstance(params["chunk_size"], int) or isinstance(params["chunk_size"], bool):
                raise TypeError(f"params: chunk_size must be an integer - the params entered is not in this required form - {params}")

            # check that chunk_size is greater than CHUNK_SIZE_MIN
            if params["chunk_size"] <= CHUNK_SIZE_MIN:
                raise ValueError(
                    f"params: chunk_size must be greater than {CHUNK_SIZE_MIN} - the params entered is not in this required form - {params}"
                )

            # check that chunk_size is no greater than CHUNK_SIZE_MAX
            if params["chunk_size"] > CHUNK_SIZE_MAX:
                raise ValueError(
                    f"params: chunk_size must be no greater than {CHUNK_SIZE_MAX} - the params entered is not in this required form - {params}"
                )

        else:
            raise ValueError(f"params: chunk_size must be in params - the params entered is not in this required form - {params}")

        # check that overlap_size is an integer
        if "overlap_size" in params.keys():
            if not isinstance(params["overlap_size"], int) or isinstance(params["overlap_size"], bool):
                raise TypeError(f"params: overlap_size must be an integer - the params entered is not in this required form - {params}")

            # check that overlap_size is greater than or equal to OVERLAP_SIZE_MIN
            if params["overlap_size"] < OVERLAP_SIZE_MIN:
                raise ValueError(
                    f"params: overlap_size must be greater than or equal to {OVERLAP_SIZE_MIN} - the params entered is not in this required form - {params}"
                )

            # check that overlap_size is no greater than chunk_size - 1
            if params["overlap_size"] > params["chunk_size"] - 1:
                raise ValueError(
                    f"params: overlap_size must be no greater than chunk_size - 1 - the params entered is not in this required form - {params}"
                )
        else:
            raise ValueError(f"params: overlap_size must be in params - the params entered is not in this required form - {params}")

        # check that if chunk_size or overlap_size is in params, then both are in params
        if "chunk_size" in params.keys() and "overlap_size" not in params.keys():
            raise ValueError(
                f"params: if chunk_size is in params, then overlap_size must also be in params - the params entered is not in this required form - {params}"
            )

        if "overlap_size" in params.keys() and "chunk_size" not in params.keys():
            raise ValueError(
                f"params: if overlap_size is in params, then chunk_size must also be in params - the params entered is not in this required form - {params}"
            )



import yaml
from typing import Optional, List
from collections import OrderedDict
from krixik.pipeline_builder.module import Module
from krixik.utilities.validators.data.utilities.decorators import datatype_validator
from krixik.utilities.validators.data.utilities.read_config import check_inverse_config
from krixik.pipeline_builder.utilities.config_checker import config_check
from krixik.pipeline_builder.utilities.name_checker import name_check
from krixik.pipeline_builder.utilities.chain_checker import chain_check
from krixik.pipeline_builder.utilities.savepath_checker import savepath_check
from krixik.pipeline_builder.utilities.input_checker import input_check
from krixik.pipeline_builder import MAX_MODULES


def represent_ordereddict(dumper, data):
    return dumper.represent_dict(data.items())


def construct_ordereddict(loader, node):
    return OrderedDict(loader.construct_pairs(node))


yaml.add_representer(OrderedDict, represent_ordereddict)
yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_ordereddict)


def convert_to_dict(obj):
    if isinstance(obj, OrderedDict):
        return {key: convert_to_dict(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [convert_to_dict(item) for item in obj]
    else:
        return obj


class CreatePipeline:
    def __init__(
        self,
        name: Optional[str] = None,
        module_chain: Optional[List[Module]] = None,
        config_path: Optional[str] = None,
    ) -> None:
        self.name = name
        self.__module_chain = []
        self.__module_chain_names = []
        self.__module_chain_output_process_keys = []
        self.__pipeline_config = None
        self.__module_chain_configs = []

        if self.name is not None:
            name_check(self.name)

        if config_path is not None:
            self.load(config_path)

        if module_chain is not None:
            chain_check(module_chain)
            for module in module_chain:
                self._add(module)
            self.test_connections()

    def _add(self, module: Module, insert_index: int = -1) -> None:
        if len(self.__module_chain) + 1 > MAX_MODULES:
            raise ValueError(f"cannot add additional module - pipelines cannot currently have more than {MAX_MODULES} modules")
        if not isinstance(module, Module):
            raise TypeError(f"FAILURE: module - {module} - is not a proper Module object")

        if insert_index != -1:
            if insert_index < -1 or insert_index > len(self.__module_chain):
                raise Exception("insert_index out of range")

            self.__module_chain.insert(insert_index, module)
            self.__module_chain_names.insert(insert_index, module.name)
            self.__module_chain_configs.insert(insert_index, module.config)
            self.__module_chain_output_process_keys.insert(insert_index, module.output_process_key)

        self.__module_chain.append(module)
        self.__module_chain_names.append(module.name)
        self.__module_chain_configs.append(module.config)
        self.__module_chain_output_process_keys.append(module.output_process_key)

        self.test_connections()

    def _remove(
        self,
        module_name: str | None = None,
        index: int | None = None,
    ) -> None:
        if index is not None and (module_name is not None):
            raise Exception("index and module_name cannot both be not None")

        if len(self.__module_chain) == 0:
            raise Exception("no modules to remove")

        # try to find index of module_custom_name in module_chain_custom_names
        if module_name is not None:
            if module_name not in self.__module_chain_names:
                raise Exception(f"module {module_name} not found in pipeline")
        else:
            module_index = self.__module_chain_names.index(module_name)
            self.__module_chain.pop(module_index)
            self.__module_chain_names.pop(module_index)
            self.__module_chain_output_process_keys.pop(module_index)
            self.__module_chain_configs.pop(module_index)

        if index is not None:
            if index < 0 or index >= len(self.__module_chain):
                raise Exception("index out of range")
            self.__module_chain.pop(index)
            self.__module_chain_names.pop(index)
            self.__module_chain_output_process_keys.pop(index)
            self.__module_chain_configs.pop(index)

        # test connections
        self.test_connections()

    @datatype_validator
    def test_input(self, *, local_file_path: str) -> None:
        input_check(local_file_path, self.__module_chain)
        print(f"SUCCESS: local file {local_file_path} passed pipeline input test passed")

    @property
    def module_chain(self) -> list:
        return self.__module_chain_names

    @property
    def module_chain_output_process_keys(self) -> list:
        return self.__module_chain_output_process_keys

    def test_connections(self):
        # if only one module, then no need to check
        if len(self.module_chain) == 1:
            return

        # check if the output_process_key is prev module
        for i in range(len(self.__module_chain) - 1):
            prev_module = self.__module_chain[i]
            prev_module_output_process_key = prev_module.output_process_key
            prev_module_output_process_key_type = prev_module.output_process_type
            prev_module_output_format = prev_module.output_format

            curr_module = self.__module_chain[i + 1]
            curr_module_input_process_key = curr_module.input_process_key
            curr_module_input_process_key_type = curr_module.input_process_type
            curr_module_input_format = curr_module.input_format

            # check format compatibility
            if prev_module_output_format != curr_module_input_format:
                raise TypeError(
                    f"format type mismatch between {prev_module.name} - whose output format is {prev_module_output_format} - and {curr_module.name} - whose input format is {curr_module_input_format}"
                )

            # check process key type compatibility
            if prev_module_output_process_key_type != curr_module_input_process_key_type:
                raise TypeError(
                    f"process key type mismatch between {prev_module.name} - whose output process key is {prev_module_output_process_key} and whose type is {prev_module_output_process_key_type} - and {curr_module.name} - whose input process key {curr_module_input_process_key} whose type is {curr_module_input_process_key_type}"
                )

    def _make_config(self):
        self.__pipeline_config = OrderedDict()
        modules = []
        for ind, mm in enumerate(self.__module_chain_configs):
            module_dict = OrderedDict()
            module = mm["module"]
            module_dict["name"] = module["name"]
            module_dict["models"] = []
            for m in module["models"]:
                entry = {}
                entry["name"] = m["name"]
                if "params" in list(m.keys()):
                    entry["params"] = OrderedDict(m["params"])
                module_dict["models"].append(OrderedDict(entry))

            if "params" in list(module["defaults"].keys()):
                module_dict["defaults"] = OrderedDict({"model": module["defaults"]["model"], "params": OrderedDict(module["defaults"]["params"])})
            else:
                module_dict["defaults"] = OrderedDict({"model": module["defaults"]["model"]})

            module_dict["input"] = OrderedDict(
                {
                    "type": module["input"]["type"],
                    "permitted_extensions": module["input"]["permitted_extensions"],
                }
            )
            module_dict["output"] = OrderedDict({"type": module["output"]["type"]})
            modules.append(module_dict)

        self.__pipeline_config["pipeline"] = OrderedDict({"name": self.name, "modules": modules})

    @property
    def config(self):
        self._make_config()
        return convert_to_dict(self.__pipeline_config)

    def save(self, *, config_path: str) -> None:
        savepath_check(config_path)

        if self.name is None:
            raise ValueError("please give your pipeline a name before saving using the .name property")

        self._make_config()
        with open(config_path, "w") as file:
            yaml.dump(self.__pipeline_config, file, default_flow_style=False, sort_keys=False)

    def load(self, pipeline_config_path: str) -> None:
        config_check(pipeline_config_path)

        with open(pipeline_config_path, "r") as file:
            pipeline_config = yaml.safe_load(file)

        pipeline = pipeline_config["pipeline"]
        pipeline_name = pipeline["name"]
        modules = pipeline["modules"]

        if len(modules) > MAX_MODULES:
            raise ValueError(f"pipelines cannot currently have more than {MAX_MODULES} modules")

        module_chain = []
        for m in modules:
            module_name = m["name"]

            module = Module(
                module_type=module_name,
            )
            module_chain.append(module)

        self.__init__(name=pipeline_name, module_chain=module_chain)
        self._make_config()



import json
import functools
from typing import Callable
import requests
from krixik.system_builder.functions import cli_checkin_endpoint


def checkin(api_key: str | None, api_url: str | None) -> int:
    if api_key is None or api_url is None:
        raise ValueError("api_key and api_url cannot be None")

    # prep headers
    headers: dict[str, str] = {"Content-Type": "text/plain", "krixikApiKey": api_key}

    try:
        # make request
        response = requests.post(
            api_url + cli_checkin_endpoint,
            headers=headers,
            json=json.dumps({}),
            timeout=15,
        )

        if response.status_code == 200:
            print("SUCCESS: You are now authenticated.")
            return 0
        else:
            print(f"FAILURE: You failed to authenticate - status code {response.status_code}")
            return 1
    except requests.exceptions.HTTPError as e:
        print(f"FAILURE: You failed to authenticate - HTTPError {e}")
        return 1
    except requests.exceptions.ConnectionError as e:
        print(f"FAILURE: You failed to authenticate - ConnectionError {e}")
        return 1
    except requests.exceptions.Timeout as e:
        print(f"FAILURE: You failed to authenticate - Timeout {e}")
        return 1
    except requests.exceptions.RequestException as e:
        print(f"FAILURE: You failed to authenticate - request exception {e}")
        return 1
    except Exception as e:
        print(f"FAILURE: You failed to authenticate - exception {e}")
        return 1


def check_init(self):
    if not (hasattr(self, "_KrixikBasePipeline__api_check_val") or hasattr(self, "_KrixikSearchPipeline__api_check_val")):
        raise ValueError("you are not authenticated - call init() to authenticate using your API key and url")
    else:
        if hasattr(self, "_KrixikBasePipeline__api_check_val"):
            if self._KrixikBasePipeline__api_check_val == 1:
                raise ValueError("you are not authenticated - all init() to authenticate using your API key and url")
            if self._KrixikBasePipeline__api_check_val is None:
                raise ValueError("you are not authenticated - call init() to authenticate using your API key and url")
        if hasattr(self, "_KrixikSearchPipeline__api_check_val"):
            if self._KrixikSearchPipeline__api_check_val == 1:
                raise ValueError("you are not authenticated - all init() to authenticate using your API key and url")
            if self._KrixikSearchPipeline__api_check_val is None:
                raise ValueError("you are not authenticated - call init() to authenticate using your API key and url")


def check_init_decorator(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        check_init(self)
        return func(self, *args, **kwargs)

    return wrapper



import os
import numpy as np


def is_valid(local_file_path: str) -> None:
    try:
        # check that local_file_path is a text file
        _, file_extension = os.path.splitext(local_file_path)
        if file_extension not in [".npy"]:
            raise ValueError(f"invalid local_file_path: please check that it is an npy file - {local_file_path}")
        np_array = np.load(local_file_path, mmap_mode="r")
        partial_array = np_array[:4096]
    except UnicodeDecodeError:
        raise ValueError(f"invalid local_file_path: please check that it is an npy file - {local_file_path}")
    except FileNotFoundError:
        raise FileNotFoundError(f"the file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"npy validation failed with exception {e}")


def is_size(
    *,
    local_file_path: str | None,
    minimum_word_count: int = 6,
    maximum_line_count: int = 100000,
    minimum_file_size: float = 0.000001,
    maximum_file_size: float = 3.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"text size calculation failed with exception {e}")

    # check size of input text file
    if local_file_path is None:
        raise ValueError("invalid local_file_path - local_file_path cannot be None")
    try:
        # check that local_file_path represents a valid text file
        is_valid(local_file_path)

        # compute file size in megabytes
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size:
            raise ValueError(
                f"input file size is {file_size} megabytes: either less than {minimum_file_size} megabytes (current minimum size allowable) or greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path}"
            )
        if file_size > maximum_file_size:
            raise ValueError(f"file size is greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path}")

    except ValueError as ve:
        raise ve

    except Exception as e:
        raise ValueError(f"text extraction failed with exception {e}")



import copy
import inspect
import tempfile
from krixik.utilities.converters.read_config import convert_extension
from krixik.utilities.converters.unclean_to_clean_txt import (
    convert as convert_unclean_text,
)
from krixik.utilities.converters.docx_to_txt import convert as convert_docx
from krixik.utilities.converters.pdf_to_txt import convert as convert_pdf
from krixik.utilities.converters.pptx_to_txt import convert as convert_pptx
from krixik.utilities.converters.video_to_audio import convert as convert_video
from krixik.utilities.utilities import vprint, get_input


def datatype_converter_wrapper(func):
    def converter_wrapper(*args, **kwargs):
        try:
            signature = inspect.signature(func)
            verbose = get_input("verbose", signature, kwargs, default_value=True)
            local_file_path = get_input("local_file_path", signature, kwargs, default_value=True)

            if local_file_path is not None:
                extension = local_file_path.split(".")[-1]
                conversion = convert_extension(extension)
                if conversion is not None:
                    with tempfile.TemporaryDirectory() as conversion_save_directory:
                        og_local_file_path = copy.deepcopy(local_file_path)
                        if extension == "mp4":
                            local_file_path = convert_video(
                                local_file_path=local_file_path,
                                local_save_directory=conversion_save_directory,
                                verbose=verbose,
                            )
                        elif extension == "pdf":
                            local_file_path = convert_pdf(
                                local_file_path=local_file_path,
                                local_save_directory=conversion_save_directory,
                                verbose=verbose,
                            )
                        elif extension == "docx":
                            local_file_path = convert_docx(
                                local_file_path=local_file_path,
                                local_save_directory=conversion_save_directory,
                                verbose=verbose,
                            )
                        elif extension == "pptx":
                            local_file_path = convert_pptx(
                                local_file_path=local_file_path,
                                local_save_directory=conversion_save_directory,
                                verbose=verbose,
                            )
                        if local_file_path is not None:
                            if local_file_path.split(".")[-1] != conversion:
                                raise ValueError(f"conversion failed, expected {conversion} got {local_file_path.split('.')[-1]}")

                            vprint(
                                f"converted {og_local_file_path} to: {local_file_path}",
                                verbose=verbose,
                            )
                            # if local_file_path.split(".") == "txt":
                            #     clean_options = None
                            #     if "clean_options" in list(kwargs.keys()):
                            #         if kwargs["clean_options"] is not None:
                            #             clean_options = kwargs["clean_options"]
                            #     use_default_clean_options = False
                            #     if "use_default_clean_options" in list(kwargs.keys()):
                            #         if kwargs["use_default_clean_options"] is not None:
                            #             use_default_clean_options = kwargs[
                            #                 "use_default_clean_options"
                            #             ]
                            #     local_file_path = convert_unclean_text(
                            #         local_file_path=local_file_path,
                            #         local_save_directory=local_save_directory,
                            #         clean_options=clean_options,
                            #         use_default_clean_options=use_default_clean_options,
                            #         verbose=verbose,
                            #     )
                        kwargs["local_file_path"] = local_file_path
                        kwargs["og_local_file_path"] = og_local_file_path
                        return func(*args, **kwargs)
            return func(*args, **kwargs)
        except ValueError as e:
            raise ValueError(e)
        except TypeError as e:
            raise TypeError(e)
        except FileNotFoundError as e:
            raise FileNotFoundError(e)
        except PermissionError as e:
            raise PermissionError(e)
        except Exception as e:
            raise Exception(e)

    return converter_wrapper



from krixik.utilities.validators import Q_MIN, Q_MAX
from krixik.utilities.utilities import invalid_char_check


def query_checker(query: str) -> None:
    if query is not None:
        # check that query is a string
        if not isinstance(query, str):
            raise TypeError(f"invalid query: not a string - {query}")

        # strip query of whitespace
        query = query.strip()

        # check that query length is greater than Q_MIN
        if len(query) <= Q_MIN:
            raise ValueError(f"invalid query: length is less than {Q_MIN} characters (current minimum length allowable): {query}")

        # check that query length is less than Q_MAX
        if len(query) > Q_MAX:
            raise ValueError(f"invalid query: length is greater than {Q_MAX} characters (current maximum length allowable) - {query}")

        # # check that query contains only acceptable characters
        # invalid_char = invalid_char_check(query)

        # if len(invalid_char) > 0:
        #     raise ValueError(
        #         f"invalid query: please remove or replace the following invalid characters - {invalid_char}"
        #     )



import inspect
from functools import wraps
from krixik.modules.utilities.module_selections import (
    pipeline_selection_setup,
)
from krixik.utilities.utilities import get_input
from krixik.utilities.utilities import vprint
from krixik.modules.utilities.io_validator import is_valid as is_valid_json_input


def type_check_inputs(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            signature = inspect.signature(func)
            self_arg = args[0] if args and hasattr(args[0], "__dict__") else None

            # confirm and hydrate module chain
            pipeline_ordered_modules = self_arg.module_chain if self_arg else kwargs["pipeline_ordered_modules"]
            modules = get_input("modules", signature, kwargs)
            verbose = get_input("verbose", signature, kwargs, default_value=True)
            hydrated_modules = pipeline_selection_setup(pipeline_ordered_modules, modules)
            kwargs["modules"] = hydrated_modules

            if "local_file_path" in list(kwargs.keys()):
                local_file_path = get_input("local_file_path", signature, kwargs)
                is_valid_json_input(pipeline_ordered_modules[0], local_file_path)

            if modules != hydrated_modules:
                vprint(
                    f"INFO: hydrated input modules: {hydrated_modules}",
                    verbose=verbose,
                )

        except ValueError as e:
            raise e
        except TypeError as e:
            raise e
        except FileNotFoundError as e:
            raise e
        except PermissionError as e:
            raise e
        except Exception as e:
            raise e

        return func(self_arg, *args[1:], **kwargs)

    setattr(wrapper, "module_selection_type_checked", bool(True))
    return wrapper



from krixik.utilities.validators.system.base.symbolic_paths import (
    symbolic_file_paths_checker as symbolic_file_paths_checker_utility,
)
from krixik.utilities.validators.system.base.symbolic_paths import (
    individual_symbolic_file_path_checker as individual_symbolic_file_path_checker_utility,
)
from krixik.utilities.validators.system.data.audio import valid_extensions


def individual_symbolic_file_path_checker(file_path: str) -> None:
    return individual_symbolic_file_path_checker_utility(file_path, valid_extensions=valid_extensions)


def symbolic_file_paths_checker(symbolic_file_paths: list, verbose: bool = True) -> None:
    return symbolic_file_paths_checker_utility(symbolic_file_paths, valid_extensions=valid_extensions, verbose=verbose)



from krixik.modules.parser import module_config
from krixik.modules.parser.params import fixed_params_validator
from krixik.modules.utilities.model import base_model_setup
from krixik.modules.utilities.model_selection import ModelSelection

available_models = module_config["available_models"]
default_model = module_config["default_model"]
default_params = module_config["default_params"]


def params_validator(model: str, params: dict) -> None:
    if model == "fixed":
        fixed_params_validator(params)
    if model == "sentence":
        if params != {}:
            raise ValueError(f"the sentence parser does not take in any parameters - you entered {params}")


def model_setup(model_selection: dict):
    return base_model_setup(
        model_selection,
        available_models,
        default_model,
        default_params,
        params_validator,
    )


def model_selection_setup(data: dict) -> ModelSelection:
    return ModelSelection(data, model_setup)



import yaml
import os


def config_check(filename):
    if not os.path.exists(filename):
        raise FileExistsError(f"FAILURE: config file does not exist - {filename}")

    _, extension = os.path.splitext(filename)
    if extension.lower() == ".yaml" or extension.lower() == ".yml":
        try:
            with open(filename, "r") as file:
                yaml.safe_load(file)
        except yaml.YAMLError:
            raise yaml.YAMLError(f"FAILURE: config is not a valid yaml - {filename}")
    else:
        raise ValueError(f"FAILURE: config - {filename} - does not end with valid extension .yml or .yaml")



import yaml
import importlib
import inspect
from functools import wraps
from krixik.utilities.validators.system.base.verbose import verbose_checker
from krixik.utilities.validators.system.base.symbolic_paths import (
    individual_symbolic_directory_path_checker,
)
from krixik.__base__ import library_base_dir
from krixik.utilities.file_name_generator import generate_random_file_name
from krixik.utilities.utilities import vprint
from krixik.utilities.utilities import get_input
from krixik.utilities.validators.data.utilities.read_config import get_allowable_data_types
from krixik.modules.utilities.read_config import get_module_available_defaults
from krixik.utilities.validators.data.utilities.read_config import get_all_allowable_extensions
from krixik.utilities.validators.system.base.extension_enforcer import (
    pairwise_extension_checker,
)
from krixik.utilities.validators.system.base.symbolic_path_splitter import symbolic_path_splitter


def get_module_input_type(module_name: str) -> str:
    module_config_path = library_base_dir + f"/modules/{module_name}/module.yml"
    with open(module_config_path, "r") as file:
        module_config = yaml.safe_load(file)
        return module_config["module"]["input"]["type"]


def type_check_inputs(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            signature = inspect.signature(func)
            self_arg = args[0] if args and hasattr(args[0], "__dict__") else None

            if "local_file_path" in list(kwargs.keys()):
                local_file_path = kwargs["local_file_path"]
                allowable_extensions = get_all_allowable_extensions()
                if "." + local_file_path.split(".")[-1] not in allowable_extensions:
                    raise ValueError(
                        f"the file extension '{local_file_path.split('.')[-1]}' is not allowed - current allowable extensions are {allowable_extensions}"
                    )

            # retrieve first module of pipeline
            pipeline_ordered_modules = self_arg.module_chain if self_arg else kwargs["pipeline_ordered_modules"]

            # load in appropriate module based on input type
            first_module = pipeline_ordered_modules[0]
            module_input_type = get_module_input_type(first_module)
            file_names_module = importlib.import_module(f"krixik.utilities.validators.system.data.{module_input_type}.file_names")
            local_file_path_module = importlib.import_module(f"krixik.utilities.validators.system.data.{module_input_type}.local_file_path")
            symbolic_file_paths_module = importlib.import_module(f"krixik.utilities.validators.system.data.{module_input_type}.symbolic_file_paths")

            file_names_checker = file_names_module.file_names_checker
            individual_file_name_checker = file_names_module.individual_file_name_checker
            local_file_path_checker = local_file_path_module.local_file_path_checker

            symbolic_file_paths_checker = symbolic_file_paths_module.symbolic_file_paths_checker
            individual_symbolic_file_path_checker = symbolic_file_paths_module.individual_symbolic_file_path_checker

            verbose = get_input("verbose", signature, kwargs, default_value=True)
            verbose_checker(verbose)

            file_names = None
            file_name = None
            local_file_path = None
            symbolic_file_paths = None
            symbolic_file_path = None
            symbolic_directory_path = None

            if "file_names" in list(kwargs.keys()):
                file_names = kwargs["file_names"]
                file_names_checker(file_names)

            if "file_name" in list(kwargs.keys()):
                file_name = kwargs["file_name"]
                individual_file_name_checker(file_name)

            if "local_file_path" in list(kwargs.keys()):
                local_file_path = kwargs["local_file_path"]
                local_file_path_checker(local_file_path)

                path_to_first_module = f"{library_base_dir}/modules/{first_module}/module.yml"
                default_module_config = get_module_available_defaults(path_to_first_module)
                input_type = default_module_config["input_type"]["type"]
                allowable_data_types = get_allowable_data_types(input_type)
                if "." + local_file_path.split(".")[-1] not in allowable_data_types:
                    raise ValueError(f"the file extension '{local_file_path.split('.')[-1]}' is not allowed for this module")

            if "symbolic_file_paths" in list(kwargs.keys()):
                symbolic_file_paths = kwargs["symbolic_file_paths"]
                symbolic_file_paths_checker(symbolic_file_paths)

            if "symbolic_file_path" in list(kwargs.keys()):
                symbolic_file_path = kwargs["symbolic_file_path"]
                individual_symbolic_file_path_checker(symbolic_file_path)

            if "symbolic_directory_path" in list(kwargs.keys()):
                symbolic_directory_path = kwargs["symbolic_directory_path"]
                individual_symbolic_directory_path_checker(symbolic_directory_path)

            if (file_name is not None or symbolic_directory_path is not None) and symbolic_file_path is not None:
                raise ValueError("file_name and symbolic_directory_path cannot both be given if symbolic_file_path is given")

            if symbolic_file_path is not None:
                symbolic_directory_path, file_name = symbolic_path_splitter(symbolic_file_path)
                kwargs["symbolic_file_path"] = None
                kwargs["symbolic_directory_path"] = symbolic_directory_path
                kwargs["file_name"] = file_name

            if "file_name" in list(kwargs.keys()) and "local_file_path" in list(kwargs.keys()):
                if "og_local_file_path" in list(kwargs.keys()):
                    local_file_path = kwargs["og_local_file_path"]

                if file_name is not None and local_file_path is not None:
                    pairwise_extension_checker(file_name, local_file_path)

                if file_name is None and local_file_path is not None:
                    file_name = generate_random_file_name(local_file_path.split(".")[-1])
                    kwargs["file_name"] = file_name
                    vprint(
                        f"INFO: file_name was not set by user - setting to random file name: {file_name}",
                        verbose=verbose,
                    )

        except ValueError as e:
            raise ValueError(e)
        except TypeError as e:
            raise TypeError(e)
        except FileNotFoundError as e:
            raise FileNotFoundError(e)
        except PermissionError as e:
            raise PermissionError(e)
        except Exception as e:
            raise Exception(e)

        return func(self_arg, *args[1:], **kwargs)

    setattr(wrapper, "system_data_type_checked", bool(True))
    return wrapper



import re
from collections import Counter
from krixik.utilities.validators import MAX_CLEAN_OPTIONS_COUNT
from krixik.utilities.validators import (
    CLEAN_OPTION_VALUE_MAX_LENGTH,
)


# validity pattern for clean_option key and value
tags_pattern = r'^[a-zA-Z0-9\s\-_.,;\'"!?]*$'


def contains_only_alphanumeric(input_string: str) -> bool:
    return bool(re.match(tags_pattern, input_string))


# check individual clean option
def individual_clean_option_checker(key: str, value: str) -> None:
    # check that key is a string
    if not isinstance(key, str):
        raise TypeError(f"invalid clean_option key: not a string - {key}")

    # check that key length is greater than 0
    if len(key) != 1:
        raise ValueError(f"invalid clean_option key: length is not equal to 1 - {key}")

    # check that key is not equal to ' '
    if key == " ":
        raise ValueError(f"invalid clean_option key: key cannot be a space - {key}")

    # check that value is a string
    if not isinstance(value, str):
        raise TypeError(f"invalid clean_option value: not a string - {value}")

    # check that value length is less than CLEAN_OPTION_VALUE_MAX_LENGTH
    if len(value) > CLEAN_OPTION_VALUE_MAX_LENGTH:
        raise ValueError(f"invalid clean_option value: length is greater than 32 (current maximum length allowable) - {value}")

    # check that value contains only alphanumeric and alphanumeric characters
    res = contains_only_alphanumeric(value)
    if res is False:
        raise ValueError(f"invalid clean_option value: please check that it contains only basic alphanumeric characters or punctuation - {value}")

    # check that key is not equal to value
    if key == value:
        raise ValueError(f"invalid clean_option: key and value cannot be the same - {key} and {value}")


# check clean_options
def clean_options_checker(clean_options: dict) -> None:
    if clean_options is not None:
        # check that clean_options is a dict
        if not isinstance(clean_options, dict):
            raise TypeError("clean_options is not a dict")

        # if dict is empty, raise error
        if len(clean_options) == 0:
            raise ValueError("invalid clean_options: empty dict provided, you must provide at least one clean_option")

        # check that clean_options is less or equal to MAX_CLEAN_OPTIONS_COUNT
        if len(clean_options) > MAX_CLEAN_OPTIONS_COUNT:
            raise ValueError(
                f"invalid clean_options: more than {MAX_CLEAN_OPTIONS_COUNT} clean_options provided, you can have a maximum of {MAX_CLEAN_OPTIONS_COUNT}"
            )

        # loop over keys / values and check for validity
        keys = list(clean_options.keys())
        values = list(clean_options.values())
        for key, value in zip(keys, values):
            individual_clean_option_checker(key, value)

        # check that keys and values share no common elements
        common_elements = set(keys) & set(values)
        if len(common_elements) > 0:
            raise ValueError(f"invalid clean_options: keys and values cannot share common elements - {common_elements}")

        # check for duplicates - print duplicates
        c = Counter(keys)
        invalid_keys = []
        for item in c.items():
            key = item[0]
            value = item[1]
            if value > 1:
                invalid_keys.append(key)

        # print invalid keys
        if len(invalid_keys) > 0:
            raise ValueError(f"invalid clean_options: duplicate keys found: {invalid_keys}")



from typing import Dict, Any, Callable
from dataclasses import dataclass


@dataclass
class ModelSelection:
    model_selection: Dict[Dict[str, str], Dict[str, Any]]
    model_setup: Callable

    def __post_init__(self):
        self.setup_result = self.model_setup(self.model_selection)

    def get_setup_result(self):
        return self.setup_result



def savepath_check(save_path: str) -> None:
    if not isinstance(save_path, str):
        raise ValueError(f"FAILURE: config_path - {save_path} is not a string")

    try:
        file_path_ext = save_path.split(".")[-1]
    except Exception:
        raise ValueError(f"config_path - {save_path} - has no extension")

    if file_path_ext != "yml" and file_path_ext != "yaml":
        raise ValueError(f"config_path -{save_path} - must have a .yml or .yaml extension")



from dataclasses import dataclass
from typing import Any, Literal, Optional


@dataclass
class InputStructure:
    format: Literal["text"] = "text"
    filename: str = "input_text.txt"
    process_key: None = None

    @property
    def data_example(self):
        return "sample text looks like this."

    @property
    def process_type(self):
        return str(self.__annotations__[self.process_key]) if self.process_key is not None else None


@dataclass
class OutputStructure:
    format: Literal["text"] = "text"
    filename: str = "input_text.txt"
    process_key: None = None

    @property
    def data_example(self):
        return "sample text looks like this."

    @property
    def process_type(self):
        return str(self.__annotations__[self.process_key]) if self.process_key is not None else None



from os import listdir
from os.path import isfile, join
from krixik.__base__ import library_base_dir

data_directory = library_base_dir + "/utilities/validators/data"
not_data = [
    "__pycache__",
    "utilities",
    "__init__.py",
    "read_config.py",
    "config.yml",
    "decorators.py",
]
available_data_types = [
    name.split(".")[0]
    for name in [f for f in listdir(data_directory) if isfile(join(data_directory, f))]
    if name not in not_data and len(name.split(".")[0]) > 0
]



import inspect
from typing import Any
from krixik.utilities.converters.default_clean_options import acceptable_chars


class classproperty:
    def __init__(self, func):
        self.fget = func

    def __get__(self, instance, owner):
        return self.fget(owner)


def invalid_char_check(line: str) -> set[str]:
    invalid_char: set[str] = set()
    for char in line:
        if ord(char) > 128 and char not in acceptable_chars:
            invalid_char.add(char)
    return invalid_char


def vprint(message: str, *, verbose: bool = False) -> None:
    # check that message is string
    if not isinstance(message, str):
        raise TypeError("message must be a string")

    # check that verbose is boolean
    if not isinstance(verbose, bool):
        raise TypeError("verbose must be a boolean")

    if verbose:
        print(message)


def get_input(
    input_name: str | None,
    signature: inspect.Signature,
    kwargs: dict[str, str],
    default_value: Any = None,
) -> Any:
    inspect._empty
    input_value = signature.parameters.get(input_name).default if input_name in signature.parameters else None
    if input_value is inspect._empty:
        input_value = None
    if input_name in list(kwargs.keys()):
        input_value = kwargs[input_name]
    if input_value is None:
        input_value = default_value
    return input_value


def load_docstring(file_path):
    with open(file_path, "r") as file:
        docstring = file.read()
    return docstring



import os
from krixik.utilities.utilities import invalid_char_check
from krixik.utilities.converters.default_clean_options import default_clean_options
from krixik.utilities.utilities import vprint


def is_valid(local_file_path: str) -> None:
    try:
        # check that local_file_path is a text file
        _, file_extension = os.path.splitext(local_file_path)
        if file_extension not in [".txt"]:
            raise ValueError(f"invalid local_file_path: please check that it is a text file - {local_file_path}")

        # try to read a portion of the file content to check that it is a text file
        with open(local_file_path, "r", encoding="utf-8") as file:
            file.read(4096)

    except UnicodeDecodeError:
        raise ValueError(f"invalid local_file_path: please check that it is a text file - {local_file_path}")
    except FileNotFoundError:
        raise FileNotFoundError(f"the file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"text validation failed with exception {e}")


def is_size(
    *,
    local_file_path: str | None,
    minimum_word_count: int = 6,
    maximum_line_count: int = 100000,
    minimum_file_size: float = 0.000001,
    maximum_file_size: float = 2.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"text size calculation failed with exception {e}")

    # compute word and line count
    def compute_word_line_count(file_path: str):
        try:
            word_count = 0
            line_count = 0
            with open(file_path, "r", encoding="utf-8") as file:
                for line in file:
                    words = line.split()
                    word_count += len(words)
                    line_count += 1
            return word_count, line_count
        except Exception as e:
            raise ValueError(f"text word and line count calculation failed with exception {e}")

    # check size of input text file
    if local_file_path is None:
        raise ValueError("invalid local_file_path - local_file_path cannot be None")
    try:
        # check that local_file_path represents a valid text file
        is_valid(local_file_path)

        # compute file size in megabytes
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size or file_size > maximum_file_size:
            raise ValueError(
                f"input file size is {file_size} megabytes: either less than {minimum_file_size} megabytes (current minimum size allowable) or greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path}"
            )

        # compute word count
        file_word_count, file_line_count = compute_word_line_count(local_file_path)

        # check that word count is greater than minimum_word_count
        if file_word_count < minimum_word_count:
            raise ValueError(f"it appears that the word count is less than the allowed minimum {minimum_word_count} words")

        # check that line count is less than maximum_line_count
        if file_line_count > maximum_line_count:
            raise ValueError(f"it appears that the line count is greater than the allowed maximum {maximum_line_count} lines")

    except ValueError as ve:
        raise ve

    except Exception as e:
        raise ValueError(f"text extraction failed with exception {e}")


def is_clean(
    *,
    local_file_path: str | None,
    minimum_word_count: int = 10,
    maximum_line_count: int = 100000,
    minimum_file_size: float = 0.000001,
    maximum_file_size: float = 3.000001,
    verbose: bool = True,
) -> dict | None:
    def check_char_ord(file_path: str | None) -> dict | None:
        if file_path is None:
            raise ValueError("invalid file_path - file_path cannot be None")

        invalid_char = set()
        with open(file_path, "r", encoding="utf-8") as file:
            for line in file:
                line_invalid_chars = invalid_char_check(line)
                invalid_char = invalid_char.union(line_invalid_chars)

        invalid_char_list = list(invalid_char)
        invalid_char_ord = [ord(char) for char in invalid_char]
        invalid_char = sorted(invalid_char, key=lambda x: invalid_char_ord[invalid_char_list.index(x)])
        clean_options = {char: str("your alphanumeric vprintable replacement for " + "'" + char[0]) for char in invalid_char}
        if len(invalid_char) > 0:
            vprint(
                f"The following non-alpha-numeric characters --> {invalid_char} <-- were found in the file: {file_path}",
                verbose=verbose,
            )
            vprint(
                "INFO: Please clean out these characters using the .process clean_options parameter and try again",
                verbose=verbose,
            )
            vprint(
                f"INFO: this argument is a dictionary that should look like the following: clean_options={clean_options}",
                verbose=verbose,
            )
            vprint(
                "INFO: do not forget to include a valid local_save_directory with this .process call to save the cleaned file",
                verbose=verbose,
            )
            vprint(
                'INFO: a list of all detected invalid characters will be returned as a dictionary with the key "invalid_characters"',
                verbose=verbose,
            )
            vprint(
                "INFO: a set of simple default clean options is available by setting use_default_clean_options=True",
                verbose=verbose,
            )
            vprint(
                "INFO: the complete set of default options is: {default_clean_options}",
                verbose=verbose,
            )
            vprint(
                "------------------------------------------------------------------------------------------------------------------",
                verbose=verbose,
            )

            default_replacements = {}
            leftovers = []
            for v in invalid_char:
                for d in default_clean_options:
                    if v in d:
                        default_replacements[v] = default_clean_options[d]
                        break
                else:
                    leftovers.append(v)

            vprint(
                "INFO: setting use_default_clean_options=True will replace the following characters in your text with the following replacements:",
                verbose=verbose,
            )
            vprint(f"INFO: {default_replacements}", verbose=verbose)

            if len(leftovers) > 0:
                vprint(
                    "INFO: the following characters will be left as is and will need suitable replacement:",
                    verbose=verbose,
                )
                vprint(f"INFO: {leftovers}", verbose=verbose)

            # package invalid_char
            invalid_char = {"invalid_characters": invalid_char}
            return invalid_char
        else:
            return None

    # check that local_file_path is valid and is less than maximum_file_size
    is_size(
        local_file_path=local_file_path,
        minimum_word_count=minimum_word_count,
        maximum_line_count=maximum_line_count,
        minimum_file_size=minimum_file_size,
        maximum_file_size=maximum_file_size,
    )

    # check for invalid characters
    ord_vals = check_char_ord(local_file_path)

    if ord_vals is not None:
        return ord_vals
    else:
        vprint(
            "SUCCESS: File check complete, file contains all valid characters",
            verbose=verbose,
        )
        return None



def pairwise_extension_checker(file_name: str, local_file_path: str) -> None:
    # check that file_name extension matches local_file_name extension
    if "." + file_name.split(".")[-1] != "." + local_file_path.split(".")[-1]:
        raise ValueError(
            f'input file_name extension must match input local_file_path extension - INFO: input file_name extension: {"." + file_name.split(".")[-1]}, input local_file_path extension: {local_file_path[-4:]}'
        )



import inspect
from functools import wraps
from krixik.utilities.utilities import vprint
from krixik.utilities.validators.system.base.verbose import verbose_checker
from krixik.utilities.utilities import get_input
from typing import Callable


# lower case an individual file_name
def lower_case_file_name(*, file_name: str, verbose: bool) -> str:
    if file_name is not None:
        lower_case_name = file_name.lower()
        if lower_case_name != file_name:
            vprint(
                f"INFO: lower casing file_name {file_name} to {lower_case_name}",
                verbose=verbose,
            )
        file_name = lower_case_name
    return file_name


# lower-case file_names
def lower_case_file_names(*, file_names: list, verbose: bool) -> list:
    if file_names is not None:
        # lower case
        lower_case_names = [lower_case_file_name(file_name=v, verbose=verbose) for v in file_names]

        # replace
        file_names = lower_case_names
    return file_names


# lower-case symbolic_directory_path
def lower_case_symbolic_directory_path(*, symbolic_directory_path: str, verbose: bool) -> str:
    if symbolic_directory_path is not None:
        lower_case_name = symbolic_directory_path.lower()
        if lower_case_name != symbolic_directory_path:
            vprint(
                f"INFO: lower casing symbolic_directory_path {symbolic_directory_path} to {lower_case_name}",
                verbose=verbose,
            )
        symbolic_directory_path = lower_case_name
    return symbolic_directory_path


# lower-case symbolic_directory_paths
def lower_case_symbolic_directory_paths(*, symbolic_directory_paths: list, verbose: bool) -> list:
    if symbolic_directory_paths is not None:
        # lower case
        lower_case_names = [lower_case_symbolic_directory_path(symbolic_directory_path=v, verbose=verbose) for v in symbolic_directory_paths]

        # replace
        symbolic_directory_paths = lower_case_names
    return symbolic_directory_paths


# lower-case symbolic_file_path
def lower_case_symbolic_file_path(*, symbolic_file_path: str, verbose: bool) -> str:
    if symbolic_file_path is not None:
        lower_case_name = symbolic_file_path.lower()
        if lower_case_name != symbolic_file_path:
            vprint(
                f"INFO: lower casing symbolic_file_path {symbolic_file_path} to {lower_case_name}",
                verbose=verbose,
            )
        symbolic_file_path = lower_case_name
    return symbolic_file_path


# lower-case symbolic_file_paths
def lower_case_symbolic_file_paths(*, symbolic_file_paths: list, verbose: bool) -> list:
    if symbolic_file_paths is not None:
        # lower case
        lower_case_names = [lower_case_symbolic_file_path(symbolic_file_path=v, verbose=verbose) for v in symbolic_file_paths]

        # replace
        symbolic_file_paths = lower_case_names
    return symbolic_file_paths


# lower case file_tags
def lower_case_file_tags(*, file_tags: list, verbose: bool) -> list:
    if file_tags is not None:
        # lower case all tags
        lower_case_tags = []
        for file_tag in file_tags:
            key = list(file_tag.keys())[0]
            value = file_tag[key]
            lower_file_tag = {key.lower(): value.lower()}
            lower_case_tags.append(lower_file_tag)

        # loop over and compare
        for lower_tag, or_tag in zip(lower_case_tags, file_tags):
            if lower_tag != or_tag:
                vprint(
                    f"INFO: lower casing file tag {or_tag} to {lower_tag}",
                    verbose=verbose,
                )

        # set file_tags to lower_case_tags
        file_tags = lower_case_tags
    return file_tags


def lower_case_decorator(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        # get name of function being called
        signature = inspect.signature(func)

        # Check each argument based on its name
        try:
            verbose = get_input("verbose", signature, kwargs, default_value=True)

            verbose_checker(verbose)

            # instantiate query args
            file_names = None
            file_name = None
            symbolic_directory_paths = None
            symbolic_directory_path = None
            symbolic_file_paths = None
            symbolic_file_path = None
            file_tags = None

            # loop over and lower-case required query args
            for arg_name, arg_value in kwargs.items():
                if arg_name.startswith("_"):
                    continue  # Skip special parameters (e.g., '_self', '_cls')
                if arg_name == "file_names":
                    file_names = lower_case_file_names(file_names=arg_value, verbose=verbose)
                    kwargs["file_names"] = file_names
                if arg_name == "file_name":
                    file_name = lower_case_file_name(file_name=arg_value, verbose=verbose)
                    kwargs["file_name"] = file_name
                if arg_name == "symbolic_directory_paths":
                    symbolic_directory_paths = lower_case_symbolic_directory_paths(symbolic_directory_paths=arg_value, verbose=verbose)
                    kwargs["symbolic_directory_paths"] = symbolic_directory_paths
                if arg_name == "symbolic_directory_path":
                    symbolic_directory_path = lower_case_symbolic_directory_path(symbolic_directory_path=arg_value, verbose=verbose)
                    kwargs["symbolic_directory_path"] = symbolic_directory_path
                if arg_name == "symbolic_file_paths":
                    symbolic_file_paths = lower_case_symbolic_file_paths(symbolic_file_paths=arg_value, verbose=verbose)
                    kwargs["symbolic_file_paths"] = symbolic_file_paths
                if arg_name == "symbolic_file_path":
                    symbolic_file_path = lower_case_symbolic_file_path(symbolic_file_path=arg_value, verbose=verbose)
                    kwargs["symbolic_file_path"] = symbolic_file_path
                if arg_name == "file_tags":
                    file_tags = lower_case_file_tags(file_tags=arg_value, verbose=verbose)
                    kwargs["file_tags"] = file_tags

        except ValueError as e:
            raise ValueError(e)
        except TypeError as e:
            raise TypeError(e)
        except FileNotFoundError as e:
            raise FileNotFoundError(e)
        except PermissionError as e:
            raise PermissionError(e)
        except Exception as e:
            raise Exception(e)

        return func(*args, **kwargs)

    setattr(wrapper, "lower_cased", bool(True))
    return wrapper



from typing import List


def get_all_key_paths(dictionary: dict, parent_key: str = "", separator: str = ".") -> List[str]:
    key_paths = []
    for key, value in dictionary.items():
        current_key = f"{parent_key}{separator}{key}" if parent_key else key
        if isinstance(value, dict):
            key_paths.extend(get_all_key_paths(value, current_key, separator))
        else:
            key_paths.append(current_key)
    return key_paths


def find_value_by_path(dictionary: dict, target_path: str, separator=".") -> dict | None:
    keys = target_path.split(separator)
    current = dictionary
    for key in keys:
        if key in current:
            current = current[key]
        else:
            return None
    return current


def find_key_path_with_type(nested_dict: dict, target_key: str, current_path=None) -> list | None:
    if current_path is None:
        current_path = []

    for key, value in nested_dict.items():
        current_path.append((key, type(value)))  # Store key and its type
        if key == target_key:
            return current_path.copy()
        if isinstance(value, dict):
            found_path = find_key_path_with_type(value, target_key, current_path)
            if found_path:
                return found_path
        current_path.pop()
    return None


def find_key_paths_with_prefix_with_type(nested_dict: dict, target_prefix: str, current_path=None) -> dict | None:
    if current_path is None:
        current_path = []

    paths = []
    for key, value in nested_dict.items():
        current_path.append((key, type(value)))  # Store key and its type
        if key.startswith(target_prefix):
            paths.append(current_path.copy())
        if isinstance(value, dict):
            found_paths = find_key_paths_with_prefix(value, target_prefix, current_path)
            paths.extend(found_paths)
        current_path.pop()
    return paths



from setuptools import setup
from setuptools import find_packages

setup(
    name="krixik",
    version="1.1.15",
    description="Easily assemble and serverlessly consume modular AI pipelines through secure Python APIs.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Jeremy Watt",
    email="jeremy@krixik.com",
    url="https://github.com/krixik-ai/krixik-cli",
    packages=find_packages(include=["krixik*"]),
    package_data={"": ["*.yml", "*.yaml"]},
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=[
        "boto3",
        "ffmpeg-python",
        "moviepy",
        "nltk",
        "numpy",
        "pillow",
        "pyaml",
        "pydub",
        "pypdf",
        "pytest",
        "pytest-subtests",
        "python-docx",
        "python-dotenv",
        "python-pptx",
        "requests",
    ],
    extras_require={
        "linting": [
            "ruff",
        ],
        "testing": [
            "pytest",
            "pytest-xdist",
            "boto3",
            "ffmpeg-python",
            "moviepy",
            "nltk",
            "numpy",
            "pillow",
            "pyaml",
            "pydub",
            "pypdf",
            "pytest",
            "pytest-subtests",
            "python-docx",
            "python-dotenv",
            "python-pptx",
            "requests",
        ],
    },
)



import yaml
from krixik.__base__ import library_base_dir
from krixik.modules import available_modules


def get_module_config_by_name(module_name: str) -> dict:
    if module_name not in available_modules:
        raise ValueError(f"module {module_name} not found")
    config_path = f"{library_base_dir}/modules/{module_name}/module.yml"
    return get_module_available_defaults(config_path)


def get_module_available_defaults(config_path: str) -> dict:
    with open(config_path, "r") as file:
        # Load the YAML data from the file
        module_config = yaml.safe_load(file)["module"]

    module_name = module_config["name"]
    available_models = [v["name"] for v in module_config["models"]]
    default_model = module_config["defaults"]["model"]
    input_type = module_config["input"]
    output_type = module_config["output"]
    default_params = {}
    if "params" in module_config["defaults"]:
        default_params = module_config["defaults"]["params"]

    module_config_local = {
        "module_name": module_name,
        "available_models": available_models,
        "default_model": default_model,
        "default_params": default_params,
        "input_type": input_type,
        "output_type": output_type,
    }

    return module_config_local



from krixik.utilities.validators import K_MIN
from krixik.utilities.validators import K_MAX


def k_checker(k: int | None = None) -> None:
    if k is not None:
        # check that k is an int, greater than 0, and less than or equal to 10
        if not isinstance(k, int) or isinstance(k, bool):
            raise TypeError(f"k must be an int greater than {K_MIN} and less than or equal to {K_MAX}")

        if k < K_MIN:
            raise ValueError(f"k must be an int greater than {K_MIN} and less than or equal to {K_MAX}")

        if k > K_MAX:
            raise ValueError(f"k must be an int greater than {K_MIN} and less than or equal to {K_MAX}")



import importlib
from collections import Counter
from krixik.modules import available_modules


def validate_module_selection(module_name: str, module_selections: dict) -> None:
    import_statement = f"krixik.modules.{module_name}.models"
    validator = importlib.import_module(import_statement)
    return validator.model_selection_setup(module_selections).get_setup_result()


def pipeline_selection_setup(pipeline_ordered_modules: list, module_selections: dict) -> dict:
    # check that pipeline_ordered_modules is a list
    if not isinstance(pipeline_ordered_modules, list):
        raise ValueError("pipeline_ordered_modules must be a list")

    # check that entries of pipeline_ordered_modules are strings
    for entry in pipeline_ordered_modules:
        if not isinstance(entry, str):
            raise TypeError("all entries of pipeline_ordered_modules must be strings")

    # check that module_selections is a dictionary
    if not isinstance(module_selections, dict):
        raise TypeError("module_selections must be a dictionary")

    module_selection_names = list(module_selections.keys())

    # check that each value of module_selections is a dict
    for key in module_selection_names:
        if not isinstance(module_selections[key], dict):
            raise TypeError("each value of module_selections must be a dict")

    # check that there are not more module_selections than there are modules in pipeline
    if len(module_selection_names) > len(pipeline_ordered_modules):
        raise ValueError(
            f"you have entered {len(module_selection_names)} module configurations but your pipeline contains only {len(pipeline_ordered_modules)} modules"
        )

    # make sure not multiple selection labels
    select_counts = Counter(module_selection_names)
    select_has_dups = any(count > 1 for count in select_counts.values())
    if select_has_dups:
        raise ValueError(f"your module selection labels cannot contain duplicate entries - {module_selection_names}")

    # pre-check that selection labels lie in available span
    module_i_labels = [f"module_{i}" for i in range(1, len(pipeline_ordered_modules) + 1)]
    available_selection_labels = available_modules + module_i_labels
    for key in module_selection_names:
        if key not in available_selection_labels:
            raise ValueError(
                f"module_selections module - {key} -  not in pipeline_ordered_modules - {pipeline_ordered_modules} - or in optional index set labels: {module_i_labels}"
            )

    # check selection names obey switching
    pipeilne_counts = Counter(pipeline_ordered_modules)
    pipeline_has_dups = any(count > 1 for count in pipeilne_counts.values())
    select_contains_index_label = any(item.startswith("module_") for item in module_selection_names)

    for key in module_selection_names:
        if pipeline_has_dups:
            if key not in module_i_labels:
                raise ValueError(
                    f"when using duplicate modules your module selection labels must lie in the index set: {module_i_labels} so that your model/param selections can be correctly mapped to the proper modules"
                )
        elif select_contains_index_label:
            if key not in module_i_labels:
                raise ValueError(
                    f"module selection names must consist of all module names or all index labels, you cannot use both - {module_selection_names}"
                )
        else:
            if key not in pipeline_ordered_modules:
                raise ValueError(f"module_selections module - {key} -  not in pipeline_ordered_modules - {pipeline_ordered_modules}")

    # check for model/param configuration for each module in pipeline
    hydrated_pipeline_selections = {}
    for module_index, module_name in enumerate(pipeline_ordered_modules):
        module_params = {}
        if module_name in module_selection_names:
            module_params = module_selections[module_name]
        if str(f"module_{module_index+1}") in module_selection_names:
            module_params = module_selections[f"module_{module_index+1}"]
        hydrated_module_selections = validate_module_selection(module_name, module_params)
        hydrated_pipeline_selections[f"module_{module_index+1}"] = hydrated_module_selections
    return hydrated_pipeline_selections



import string
import random


def generate_random_file_name(extension: str) -> str:
    if not extension:
        raise ValueError("Extension cannot be empty")
    if not isinstance(extension, str):
        raise ValueError("Extension must be a string")

    num_chars = 10
    characters = string.ascii_letters.lower()
    random_chars = "".join(random.choice(characters) for _ in range(num_chars))
    file_name = f"{random_chars}.{extension}"
    filename = "krixik_generated_file_name_" + file_name
    return filename



from functools import wraps
from krixik.utilities.validators.system.base.utilities.decorators import (
    type_check_inputs as system_base_type_check_inputs,
)
from krixik.utilities.validators.system.data.utilities.decorators import (
    type_check_inputs as system_data_type_check_inputs,
)
from krixik.utilities.validators.system.base.lower_case import lower_case_decorator
from krixik.utilities.validators.data.utilities.decorators import datatype_validator


def type_check_inputs(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            result = system_base_type_check_inputs(system_data_type_check_inputs(datatype_validator(lower_case_decorator(func))))(
                self, *args, **kwargs
            )

            return result

        except ValueError as e:
            raise ValueError(e)
        except TypeError as e:
            raise TypeError(e)
        except FileNotFoundError as e:
            raise FileNotFoundError(e)
        except PermissionError as e:
            raise PermissionError(e)
        except Exception as e:
            raise Exception(e)

    return wrapper



import os
from moviepy.editor import VideoFileClip
from krixik.utilities.validators.system.base.utilities.decorators import (
    file_converters_input_check,
)
from krixik.utilities.validators.data.video import is_size as is_video_size
from krixik.utilities.validators.data.audio import is_size as is_audio_size
from krixik.utilities.utilities import vprint


def delete_file(*, file_path: str, exception: str) -> None:
    if file_path is not None:
        if os.path.exists(file_path):
            os.remove(file_path)
        raise ValueError(f"audio extraction failed with exception: {str(exception)}")


@file_converters_input_check
def extract_audio(*, local_file_path: str, audio_filepath: str) -> None:
    try:
        video = VideoFileClip(local_file_path)
        audio = video.audio
        if audio is not None:
            audio.write_audiofile(audio_filepath, verbose=False, logger=None)
    except Exception as e:
        raise ValueError(f"error extracting audio from video {local_file_path}, exception: {e}")


@file_converters_input_check
def convert(
    *,
    local_file_path: str | None,
    local_save_directory: str | None,
    verbose: bool = True,
) -> str:
    """video to audio converter - strips audio mp3 from video file using moviepy library

    Parameters
    ----------
    local_file_path : str | None
        path to local file to convert
    local_save_directory : str | None
        local directory to save converted file
    verbose : bool, optional
        by default True

    Returns
    -------
    str | None
        if conversion successful returns path to converted file
    """

    if local_file_path is not None and local_save_directory is not None:
        audio_filepath = None
        try:
            # get file_name video_file_path
            file_name = "krixik_converted_version_" + os.path.splitext(os.path.basename(local_file_path))[0]

            # format audio file path
            extension = ".mp3"
            audio_filepath = os.path.join(local_save_directory, file_name + extension)

            # remove file if it already exists
            if os.path.exists(audio_filepath):
                os.remove(audio_filepath)

            # check input file is mp4
            is_video_size(local_file_path=local_file_path, minimum_file_size=0.1)

            # extract audio from video
            extract_audio(local_file_path=local_file_path, audio_filepath=audio_filepath)

            # report size check
            vprint(
                "INFO: Checking that file size falls within acceptable parameters...",
                verbose=verbose,
            )
            is_audio_size(local_file_path=audio_filepath, minimum_file_size=0.05)
            vprint("INFO:...success!", verbose=verbose)

            return audio_filepath
        except ValueError as ve:
            vprint(
                "INFO:...failure!  The converted (audio) file does not fall within acceptable size parameters.",
                verbose=verbose,
            )
            if audio_filepath is not None:
                delete_file(file_path=audio_filepath, exception=str(ve))
            raise
        except AttributeError as ae:
            if audio_filepath is not None:
                delete_file(file_path=audio_filepath, exception=str(ae))
            raise ValueError(f"video contains no visual content, cannot convert to audio - {local_file_path}")
        except Exception as e:
            if audio_filepath is not None:
                delete_file(file_path=audio_filepath, exception=str(e))
            raise
    else:
        raise ValueError("the input local_file_path and/or local_save_directory is null")
